#include <stdarg.h>
#include "Autopilot.h"




double orbitHeading(const double& plane_lat, const double& plane_lon, const double& wpt_lat, const double& wpt_lon, const double& wpt_hdg, const double& radius, const bool& clockwise, const double& hcMax, const double& k)
{
	double d0    = distance(plane_lat, plane_lon, wpt_lat, wpt_lon, true);
	double d     = d0 - radius;
	double theta = heading(plane_lat, plane_lon, wpt_lat, wpt_lon);
	double hc    = -(2 / M_PI) * hcMax * atan(k * d);

	if (clockwise)
	{
		theta = fmod(theta - 90, 360);
		hc *= -1;
	}
	else
		theta = fmod(theta + 90, 360);

	return fmod(hc + theta + 360, 360);
}




double linePathHeading(const double& plane_lat, const double& plane_lon, const double& wpt_lat, const double& wpt_lon, const double& wpt_hdg, const double& hcMax, const double& k)
{
	double h1    = wpt_hdg;
	double h01   = heading(plane_lat, plane_lon, wpt_lat, wpt_lon);
	double theta = fmod(h01 - h1 + 360, 360);
	double d0    = distance(plane_lat, plane_lon, wpt_lat, wpt_lon, true);
	double dPerp = d0 * tan(radians(theta));
	double hc    = -(2 / M_PI) * hcMax * atan(k * dPerp);

	return fmod(h1 - hc + 360, 360);
}




double acuteAngle(const double& refAngle, const double& angle2)
{
	double normRefAngle = 180;
	double normAngle2   = fmod(angle2 + 180 - refAngle + 360, 360);

	return fmod(normAngle2 - normRefAngle, 360);
}




double slope(const double& xi, const double& yi, const double& xf, const double& yf)
{
	return (yf - yi) / (xf - xi);
}




float float_constrain(const float& input, const float& min, const float& max)
{
	if (input > max)
		return max;
	else if (input < min)
		return min;
	else
		return input;
}




double double_constrain(const double& input, const double& min, const double& max)
{
	if (input > max)
		return max;
	else if (input < min)
		return min;
	else
		return input;
}




float float_map(const float& x, const float& in_min, const float& in_max, const float& out_min, const float& out_max)
{
	return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}




double double_map(const double& x, const double& in_min, const double& in_max, const double& out_min, const double& out_max)
{
	return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}




double heading(const double& lat_1, const double& lon_1, const double& lat_2, const double& lon_2)
{
	double deltaLon_r = radians(lon_2 - lon_1);
	double lat_1_r    = radians(lat_1);
	double lat_2_r    = radians(lat_2);

	double x = cos(lat_2_r) * sin(deltaLon_r);
	double y = cos(lat_1_r) * sin(lat_2_r) - sin(lat_1_r) * cos(lat_2_r) * cos(deltaLon_r);

	return fmod(degrees(atan2(x, y)) + 360, 360);
}




double distance(const double& lat_1, const double& lon_1, const double& lat_2, const double& lon_2, const bool& km)
{
	double lat_1_r = radians(lat_1);
	double lon_1_r = radians(lon_1);
	double lat_2_r = radians(lat_2);
	double lon_2_r = radians(lon_2);

	double deltaLat_r = lat_2_r - lat_1_r;
	double deltaLon_r = lon_2_r - lon_1_r;

	double a = (sin(deltaLat_r / 2) * sin(deltaLat_r / 2)) + cos(lat_1_r) * cos(lat_2_r) * (sin(deltaLon_r / 2)) * (sin(deltaLon_r / 2));

	if (km)
		return 2 * EARTH_RADIUS_KM * atan2(sqrt(a), sqrt(1 - a));
	else
		return 2 * EARTH_RADIUS_M * atan2(sqrt(a), sqrt(1 - a));
}




double climbAngle(const double& i_lat, const double& i_lon, const double& f_lat, const double& f_lon, const double& i_alt, const double& f_alt)
{
	double dist = distance(i_lat, i_lon, f_lat, f_lon);
	double d_alt = f_alt - i_alt;

	return degrees(atan(d_alt / dist));
}




void coord(const double& lat_1, const double& lon_1, double& lat_2, double& lon_2, const double& distance, const double& bearing, const bool& km)
{
	double bearing_r = radians(fmod(bearing + 360, 360)); // Prevent negative bearing angles
	double adj_dist;

	if (km)
		adj_dist = distance / EARTH_RADIUS_KM;
	else
		adj_dist = distance / EARTH_RADIUS_M;

	double lat_1_r = radians(lat_1);
	double lon_1_r = radians(lon_1);

	lat_2 = asin(sin(lat_1_r) * cos(adj_dist) + cos(lat_1_r) * sin(adj_dist) * cos(bearing_r));
	lon_2 = lon_1_r + atan2(sin(bearing_r) * sin(adj_dist) * cos(lat_1_r), cos(adj_dist) - sin(lat_1_r) * sin(lat_2));

	lat_2 = degrees(lat_2);
	lon_2 = degrees(lon_2);
}




void toXY(const double& lat, const double& lon, double& x, double& y, const double& refLat, const double& refLon)
{
	double dist = distance(refLat, refLon, lat, lon);
	double hdg  = heading(refLat, refLon, lat, lon);

	x = dist * cos(radians(hdg - 90));
	y = dist * sin(radians(hdg - 90));
}




double distXY(const double& x_1, const double& y_1, const double& x_2, const double& y_2)
{
	return sqrt((x_1 * x_2) + (y_1 * y_2));
}




double hdgXY(const double& x_1, const double& y_1, const double& x_2, const double& y_2)
{
	return ((y_2 - y_1) / (x_2 - x_1));
}




void toDD(const double& x, const double& y, double& lat, double& lon, const double& refLat, const double& refLon)
{
	coord(refLat, refLon, lat, lon, distXY(x, y), hdgXY(0, 0, x, y)); // Assuming 1 xy unit is 1m
}
