#pragma once
#include <stdarg.h>
#include "Arduino.h"
#include "ArduPID.h"
#include "FireTimer.h"




const double EARTH_RADIUS_M  = 6378100.0;
const double EARTH_RADIUS_KM = EARTH_RADIUS_M / 1000;
const double EARTH_GRAVITY   = 9.80665;




double orbitHeading(const double& plane_lat, const double& plane_lon, const double& wpt_lat, const double& wpt_lon, const double& wpt_hdg, const double& radius = 1, const bool& clockwise = true, const double& hcMax = 90, const double& k = 1);
double linePathHeading(const double& plane_lat, const double& plane_lon, const double& wpt_lat, const double& wpt_lon, const double& wpt_hdg, const double& hcMax = 90, const double& k = 1);
double acuteAngle(const double& refAngle, const double& angle2);
double slope(const double& xi, const double& yi, const double& xf, const double& yf);
float float_constrain(const float& input, const float& min, const float& max);
double double_constrain(const double& input, const double& min, const double& max);
float float_map(const float& x, const float& in_min, const float& in_max, const float& out_min, const float& out_max);
double double_map(const double& x, const double& in_min, const double& in_max, const double& out_min, const double& out_max);
double heading(const double& lat_1, const double& lon_1, const double& lat_2, const double& lon_2);
double distance(const double& lat_1, const double& lon_1, const double& lat_2, const double& lon_2, const bool& km = false);
double climbAngle(const double& i_lat, const double& i_lon, const double& f_lat, const double& f_lon, const double& i_alt, const double& f_alt);
void coord(const double& lat, const double& lon, double& lat_2, double& lon_2, const double& distance, const double& bearing, const bool& km = false);
void toXY(const double& lat, const double& lon, double& x, double& y, const double& refLat = 0, const double& refLon = 0);
double distXY(const double& x_1, const double& y_1, const double& x_2 = 0, const double& y_2 = 0);
double hdgXY(const double& x_1, const double& y_1, const double& x_2, const double& y_2);
void toDD(const double& x, const double& y, double& lat, double& lon, const double& refLat = 0, const double& refLon = 0);




struct __attribute__((packed)) control_params {
	double setpoint;   // Degrees
	double kp;         // Unitless
	double ki;         // Unitless
	double kd;         // Unitless
	double sampleRate; // Hz
	double outputMax;  // Unitless
	double outputMin;  // Unitless
};

struct __attribute__((packed)) pilsim_state_params {
	double roll;  //  (+ = right, - = left)
	double pitch; //  (+ = up,    - = down)
	double hdg;   // 
	double alt;   // m
	double lat;   // 
	double lon;   // 
	double ias;   // m/s
	double flaps; // % (0% = up, 100% = down)
	double gear;  // % (0% = up, 100% = down)
};

struct __attribute__((packed)) state_params {
	double roll;  //  (+ = right, - = left)
	double pitch; //  (+ = up,    - = down)

	double hdg;      // 
	double hdg_comp; // 
	double hdg_imu;  // 

	double cog;          // 
	double cog_gps;      // 
	double cog_gps_calc; // 

	double alt;       // m
	double alt_gps;   // m
	double alt_baro;  // m
	double alt_lidar; // m

	double lat;      // 
	double lon;      // 
	double prev_lat; // 
	double prev_lon; // 

	uint16_t UTC_year;
	uint8_t UTC_month;
	uint8_t UTC_day;
	uint8_t UTC_hour;
	uint8_t UTC_minute;
	uint8_t UTC_second;

	double ias;          // m/s
	double ias_pitot;    // m/s
	double ias_gps;      // m/s
	double ias_gps_calc; // m/s

	bool flaps; // 0 = up, 1 = down
	bool gear;  // 0 = up, 1 = down
};

struct point {
	double maxRoll;    // 
	double minTurnRad; // m
	double hitRadius;  // m

	double alt;     // m
	double speed;   // m/s
	double heading; // 
	double lat;     // 
	double lon;     // 

	double rc_lat;  // Right Turn Circle lat 
	double rc_lon;  // Right Turn Circle lon 
	double lc_lat;  // Left Turn Circle lat 
	double lc_lon;  // Left Turn Circle lon 
	double c_lat;   // Selected Turn Circle lat 
	double c_lon;   // Selected Turn Circle lon 
	double e_lat;   // Enter/exit lat 
	double e_lon;   // Enter/exit lon 
};

enum dubin {
	LSRU, // Left Straight Right Up
	LSRD, // Left Straight Right Down
	RSLU, // Right Straight Left Up
	RSLD, // Right Straight Left Down
	RSRU, // Right Straight Right Up
	RSRD, // Right Straight Right Down
	LSLU, // Left Straight Left Up
	LSLD  // Left Straight Left Down
};

enum nav_state {
	TAKEOFF,   // Takeoff roll
	TURN_I,    // Initial turn
	STRAIGHT,  // Straight
	TURN_F,    // Final turn
	FINAL,     // Final leg
	DISENGAGED // Disengage dubin-styled navigation
};

struct nav_frame {
	dubin path; // Dubin path type
	point ni;   // Current point
	point nf;   // Next point
};




class pitch_controller : public ArduPID
{
public:
	void begin(double* _input,
		       double* _output,
		       double* _setpoint,
		       const double& _p = 0,
		       const double& _i = 0,
		       const double& _d = 0,
		       const pOn& _pOn = P_ON_E,
		       const dir& _direction = FORWARD,
		       const unsigned int& _minSamplePeriodMs = 0,
		       const double& _bias = 0,
			   const double& _rollGain = 1);
	void compute(const double& roll);
	void setRollGain(const double& _rollGain);




private:
	double rollGain;




	double roll_compensation(const double& roll);
};




class roll_controller : public ArduPID
{

};




class heading_controller : public ArduPID
{
public:
	void compute();
};




class altitude_controller : public ArduPID
{

};




class ias_controller : public ArduPID
{

};




class navigator
{
public:
	void processFrame(nav_frame& frame);
	void findMTR(point& curPoint); // Minimum Turn Radius
	void findTurnCenters(point& curPoint);
	void findPath(nav_frame& frame); 
	void findEPts(nav_frame& frame); // Enter/Exit Point locations
};
